import argparse
from pathlib import Path
from multiprocessing import Pool
from itertools import chain

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns


def parse_benchmark_file(file_path):
    """Extract runtime from a benchmark file output by snakemake

    :param file_path: path to benchmark file
    :type file_path: Path
    :return: tuple of toolname, target species, fold no,
    walltime and cpu time as timedelta object
    """
    print(f'Processing {file_path}')
    fold = file_path.stem[-1]
    target_species = file_path.parts[-2].replace('_', ' ')
    tool = file_path.parts[-3]
    if tool == 'uncalled':
        if 'index' in file_path.stem:
            tool += ' index'
            fold = None
        else:
            tool += ' map'
    df = pd.read_csv(file_path, sep='\t')
    walltime = float(df.s)

    return tool, target_species, fold,  walltime


def main(args):
    # For debugging:
    pd.options.display.width = 0
    # Hardcoded this; number of files in test set for 5 fold CV
    num_test_reads = 67098
    num_baseless_reads = 1000

    # Runtime parsing
    baseless_benchmark_files = args.benchmark_path.glob(
        '**/baseless_fold?.txt')
    deepnano_basecall_benchmark_files = args.benchmark_path.glob(
        '**/deepnano_basecall_benchmark_fold?.txt')
    deepnano_minimap_benchmark_files = args.benchmark_path.glob(
        '**/minimap_deepnano_benchmark_fold?.txt')
    guppy_benchmark_files = args.benchmark_path.glob(
        '**/guppy_benchmark_fold?.txt')
    uncalled_map_benchmark_files = args.benchmark_path.glob(
        '**/uncalled_map_benchmark_fold?.txt')
    uncalled_index_benchmark_files = args.benchmark_path.glob(
        '**/uncalled_index_benchmark.txt')
    squiggle_benchmark_files = args.benchmark_path.glob(
        '**/inference_fold?.txt')

    with Pool(args.threads) as p:
        all_files = chain(deepnano_basecall_benchmark_files,
                          deepnano_minimap_benchmark_files,
                          guppy_benchmark_files,
                          uncalled_map_benchmark_files,
                          uncalled_index_benchmark_files,
                          squiggle_benchmark_files,
                          baseless_benchmark_files)

        all_benchmarks = p.map(parse_benchmark_file, all_files)

    assert all_benchmarks, 'No files found'

    df = pd.DataFrame.from_records(all_benchmarks, columns=['tool',
                                                            'species',
                                                            'fold',
                                                            'walltime'])

    # Set species in alphabetical order
    # Remove gingivalis from dataset
    is_not_gingivalis = df['species'].apply(lambda x: False
                                            if x.find('gingivalis') > 0
                                            else True)
    df = df[is_not_gingivalis]
    all_species = df.species.unique()
    all_species.sort()

    # Add up rows where deepnano occurs twice (for mapping + basecalling)
    df = df.groupby(['tool', 'species', 'fold'], as_index=False).sum(0)

    # TODO TEMPORARY: We only ran baseless on 1000 reads to save time so we adjust for it here
    df['wall_time_per_read'] = np.where(df.tool == 'baseless',
                                        df.walltime / num_baseless_reads,
                                        df.walltime / num_test_reads)
    # df['cpu_time_per_read'] = np.where(df.tool == 'baseless',
    #                                    df.cpu_time / num_baseless_reads,
    #                                    df.cpu_time / num_test_reads)
    df.to_csv(args.out_dir / 'benchmark_runtimes.csv',
              index=False)

    g = sns.catplot(x='species', y='wall_time_per_read', data=df, hue='tool',
                    order=all_species, kind='strip', legend=False, aspect=2,
                    hue_order=['deepnano', 'uncalled map', 'guppy',
                               'squigglenet', 'baseless'], dodge=True, jitter=0)
    g.set_xticklabels(rotation=45, ha="right",
                      rotation_mode="anchor")
    g.axes[0, 0].set_ylabel('Wall time per read (s)')
    g.axes[0, 0].set_yscale('log')
    plt.tight_layout()
    plt.savefig(args.out_dir / 'benchmark_runtimes.svg')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="""Plot runtime of
     benchmarked algorithms""")
    parser.add_argument('--benchmark-path',
                        help='Path to folder that contains all benchmarks '
                             'output by the snakemake workflow. '
                             'Should contain a folder called output_16s_files',
                        required=True, type=Path)
    parser.add_argument('--out-dir',
                        help='Directory in which to save the output figure',
                        required=True, type=Path)
    parser.add_argument('--threads',
                        help='Number of threads to run in parallel',
                        required=False, type=int, default=30)

    args = parser.parse_args()
    main(args)
