import tarfile
import numpy as np
from math import ceil, floor
import tensorflow as tf
import tensorflow.keras.backend as K
import h5py
from db_building.TrainingRead import Read

from pathlib import Path
from os.path import splitext
from tempfile import TemporaryDirectory

from nns.keras_metrics_from_logits import precision, recall, binary_accuracy


class InferenceModel(object):
    """Composite model of multiple k-mer recognising NNs."""

    def __init__(self, mod_fn, batch_size=32, threshold=5, kmer_frac=0.65):
        """Construct model

        :param mod_fn: Path to tar file that contains tarred model output by compile_model.py
        :type mod_fn: str
        :param batch_size: Size of batch
        :param kmer_frac: If this fraction of target k-mers is found;
                          yield positive hit
        """
        self.load_model(mod_fn)
        self.batch_size = batch_size
        self.threshold = threshold
        # Total number of needed positive kmers to return a positive hit
        self.kmers_needed = len(self.kmers) * kmer_frac

    def predict(self, read, kmer):
        """Search the read for a single kmer

        :param read: Numpy array of split read
        :param kmer: String that indicates target kmer
        :return: True if any of the nn outputs are above threshold
        """
        model_output = self._model_dict[kmer](read)
        return np.any(model_output > self.threshold)

    def detect_species(self, read):
        """Search for species target k-mers in read

        :param read: Numpy array of split read
        :return: 1 if species found, else 0
        """
        found_kmers = 0
        species_found = 0
        for i, kmer in enumerate(self.kmers):
            if self.predict(read, kmer):
                found_kmers += 1
                if found_kmers >= self.kmers_needed:
                    # Sufficient positive kmers have been found
                    species_found = 1
                    break
            elif i - found_kmers > len(self.kmers) - self.kmers_needed:
                # Sufficient negative kmers have been found
                break
        return species_found

    def load_model(self, mod_fn):
        """Load compiled model from path to saved precompiled model

        :param mod_fn: Path to tarred file of compiled model
        :return:
        """
        with TemporaryDirectory() as td:
            with tarfile.open(mod_fn) as fh:
                fh.extractall(td)
            out_dict = {}

            def loss_fun(y_true, y_pred):  # just dummy to satisfy the stupid thing being there
                msk = np.zeros(100, dtype=bool)
                msk[50] = True
                y_pred_single = tf.boolean_mask(y_pred, msk, axis=1)
                return K.binary_crossentropy(K.cast(y_true, K.floatx()), y_pred_single, from_logits=True)
            for mn in Path(td).iterdir():
                out_dict[splitext(mn.name)[0]] = tf.keras.models.load_model(mn,
                                                                            custom_objects={'precision': precision,
                                                                                            'recall': recall,
                                                                                            'binary_accuracy': binary_accuracy,
                                                                                            'loss_fun': loss_fun})
        # Dictionary with kmer string as key and keras.Sequential as value
        self._model_dict = out_dict
        # List of kmers that the InferenceModel contains models for
        self.kmers = list(self._model_dict)
        # Length of input that should be given to each individual model
        try:
            self.input_length = self._model_dict[list(self._model_dict)[0]].layers[0].input_shape[1]
        except:
            self.input_length = None  # todo for undefined timseries length nn, find better solution
