from pathlib import Path
from time import sleep
import h5py
import numpy as np
import sqlite3
from multiprocessing import Process
from tensorflow.keras.preprocessing.sequence import pad_sequences

from db_building.TrainingRead import Read
from inference.ReadManager import ReadManager

from helper_functions import numeric_timestamp, safe_cursor


class ReadTable(object):

    def __init__(self, reads_dir, table_fn, pos_reads_dir, kmers, input_length, batch_size=32):
        """Table that keeps track of all reads in directory and what
        kmers they contain

        :param reads_dir: Directory that contains fast5 reads on which to run inference
        :param table_fn: File name of this table database
        :param pos_reads_dir: # TODO don't quite get what this is
        :param kmers: List of kmers for which to search
        :param input_length: Length of signal that should be passed to model as input
        :param batch_size: Batch size for the model
        """
        Path(table_fn).unlink(missing_ok=True)
        self.input_length = input_length
        self.batch_size = batch_size
        self.table_fn = table_fn
        self.pos_reads_dir = pos_reads_dir
        self.reads_dir = reads_dir
        self.kmers = kmers
        self.conn = sqlite3.connect(table_fn, timeout=10)  # todo longer timeout avoids locks of get_read_to_predict, but needs more elegant fix
        # self.cursor = self.conn.cursor()
        self.read_dict = {}  # todo how to remove reads after no longer necessary?

        safe_cursor(self.conn, f"""CREATE TABLE read_table (read_id text, 
            timestamp integer, {', '.join([f'{km} integer' for km in self.kmers])} integer)""", read_only=False)

    def init_table(self):
        manager_process = Process(target=ReadManager, args=(self.table_fn,
                                                            self.reads_dir,
                                                            self.pos_reads_dir,
                                                            self.kmers), name='read_manager')
        manager_process.start()
        return manager_process

    def get_read_to_predict(self, unsplit):
        """Get read and a kmer for which to scan the read file

        :return: Tuple of: (path to read fast5 file,
                 Read object for inference (split to desired input length),
                 k-mer as string for which to scna)
        """
        tup_list = safe_cursor(self.conn, "SELECT * FROM read_table ORDER BY ROWID ASC LIMIT 10")
        if len(tup_list) == 0: return None, None, None
        for tup in tup_list:
            kmi_list = [ki for ki, kb in enumerate(tup[2:]) if kb == 0]
            if not len(kmi_list): continue
            kmer_index = kmi_list[0]
            return tup[0], self.load_read(tup[0], unsplit), self.kmers[kmer_index]
        return None, None, None

    def load_read(self, fn, unsplit=False):
        if fn in self.read_dict: return self.read_dict[fn]
        with h5py.File(fn, 'r') as fh:
            if unsplit:
                self.read_dict[fn] = Read(fh, 'median').raw
            else:
                self.read_dict[fn] = Read(fh, 'median').get_split_raw_read(self.input_length)
        return self.read_dict[fn]

    def update_prediction(self, fn, kmer, pred):
        pred_num = 1 if pred else -1
        safe_cursor(self.conn, f"UPDATE read_table SET {kmer} = {pred_num}, timestamp = {numeric_timestamp()} WHERE read_id = '{fn}'", read_only=False)
