"""Script that creates library of maximally dissimilar k-mers.
Needs scrappy to be installed. And can be run from the command line

Based largely on the following paper:

Doroschak, K., Zhang, K., Queen, M. et al.
Rapid and robust assembly and decoding of molecular tags with DNA-based
nanopore signatures. Nat Commun 11, 5454 (2020).
https://doi.org/10.1038/s41467-020-19151-8
"""

import argparse
import multiprocessing
import os
from itertools import product
from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import scipy.spatial as sp, scipy.cluster.hierarchy as hc
from _ucrdtw import ucrdtw
import json

sns.set(font_scale=1.5, style="white")


def s_w(seq1, seq2, cost_fn={"match": 1, "mismatch": -1, "gap": -4}):
    '''Smith-Waterman implementation (also use this for TAing 427)'''
    nrows = len(seq1) + 1  # + 1 is to accommodate the 0th row
    ncols = len(seq2) + 1
    dp = np.zeros((nrows, ncols))

    def score():
        best_score = -np.Inf
        best_pos = (0, 0)
        for row_i in range(1, nrows):
            for col_j in range(1, ncols):
                score_ij = _score_ij(row_i, col_j)
                dp[row_i, col_j] = score_ij
                if score_ij >= best_score:
                    best_score = score_ij
                    best_pos = (row_i, col_j)
        return best_pos, best_score

    def _score_ij(i, j):
        if seq1[i - 1] == seq2[j - 1]:
            match_cost = cost_fn["match"]
        else:
            match_cost = cost_fn["mismatch"]

        up = dp[i - 1, j] + cost_fn["gap"]  # anything but diag must be a gap
        left = dp[i, j - 1] + cost_fn["gap"]
        dia = dp[i - 1, j - 1] + match_cost

        return max(up, left, dia)

    def traceback(best_pos):
        '''Ties prefer diagonal > up > left in this implementation'''
        seq1_out, seq2_out, match_str = "", "", ""
        row_i, col_j = best_pos

        # Deal with uneven sequences
        if row_i == col_j and row_i < nrows - 1 and col_j < ncols - 1:
            seq1_out = seq1[row_i:]
            seq2_out = seq2[col_j:]
            for i in range(row_i, nrows - 1):
                match_str += ":"
        if row_i != col_j and row_i < nrows - 1:
            seq1_out = seq1[row_i:]
            for i in range(row_i, nrows - 1):
                match_str += " "
                seq2_out += "-"
        if row_i != col_j and col_j < ncols - 1:
            seq2_out = seq2[col_j:]
            for i in range(col_j, ncols - 1):
                match_str += " "
                seq1_out += "-"

        # Traceback
        last_dia_s1 = 0
        last_dia_s2 = 0
        while row_i and col_j:  # end when either is 0
            up = dp[row_i - 1, col_j]
            left = dp[row_i, col_j - 1]
            dia = dp[row_i - 1, col_j - 1]

            # Case 1: diagonal
            if dia >= up and dia >= left:
                row_i -= 1
                col_j -= 1
                last_dia_s1 = row_i
                last_dia_s2 = col_j
                seq1_out = seq1[row_i] + seq1_out
                seq2_out = seq2[col_j] + seq2_out
                if seq1[row_i] == seq2[col_j]:
                    match_str = "|" + match_str
                else:
                    match_str = ":" + match_str
            # Case 2: up
            elif up >= left:
                row_i -= 1
                seq1_out = seq1[row_i] + seq1_out
                seq2_out = "-" + seq2_out
                match_str = " " + match_str
            # Case 3: left
            else:
                col_j -= 1
                seq1_out = "-" + seq1_out
                seq2_out = seq2[col_j] + seq2_out
                match_str = " " + match_str

        # Deal with uneven sequences
        if 0 < row_i:
            seq1_out = seq1[:row_i] + seq1_out
            for i in range(0, row_i):
                seq2_out = "-" + seq2_out
                match_str = " " + match_str
        if 0 < col_j:
            seq2_out = seq2[:col_j] + seq2_out
            for i in range(0, col_j):
                seq1_out = "-" + seq1_out
                match_str = " " + match_str

        return seq1_out, seq2_out, match_str, last_dia_s1, last_dia_s2

    best_pos, best_score = score()
    seq1_out, seq2_out, match_str, last_dia_s1, last_dia_s2 = traceback(
        best_pos)
    return best_pos, best_score, "\n".join([seq1_out, match_str, seq2_out]), \
           last_dia_s1, last_dia_s2


def reverse_complement(km):
    complement_dict = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
    return ''.join([complement_dict[k] for k in km][::-1])


def find_gc_content(sequence):
    gc_content = 1. * (sequence.count("G") +
                       sequence.count("C")) / len(sequence)
    return gc_content


def check_valid_seq(sequence, remove_if_outer10_percent=False):
    """We will omit the GC range limit in this simulation, originally the gc percentage had to be between 0.3 and 0.7
    """

    valid = True
    reason = []

    # Check homopolymers
    for h in HOMOPOLYMERS:
        if h in sequence:
            reason.append("Contains a homopolymer")
            valid = False
            break

    # If remove_if_outer10_percent is set to true; remove sequences that
    # have an abundance in the outer 10% of all 16s sequences
    if remove_if_outer10_percent and sequence in OUTER10_PERCENT:
        reason.append('Extreme presence or absence')
        valid = False

    return valid, reason


def get_dynamic_range(squiggle):
    dr = np.abs(np.max(squiggle) - np.min(squiggle))
    return dr


def check_valid_squig(squiggle):
    valid_dr = ("X", "Y")
    dr = get_dynamic_range(squiggle)
    if dr < valid_dr[0] or dr > valid_dr[1]:
        return False
    return True


def mutate(sequence, n_mutations=1):
    assert n_mutations <= len(sequence)
    new_sequence = list(sequence)
    nts = ("A", "C", "G", "T")
    ix_to_mutate = np.random.choice(
        range(len(new_sequence)), size=n_mutations, replace=False)
    for i in ix_to_mutate:
        possible_bases = list(nts)
        possible_bases.remove(new_sequence[i])
        new_sequence[i] = np.random.choice(possible_bases, size=1)[0]
    new_sequence = "".join([str(x) for x in new_sequence])

    # Test
    n_diff = 0
    for i in range(len(sequence)):
        if sequence[i] != new_sequence[i]:
            n_diff += 1
    assert n_diff == n_mutations
    return new_sequence


def mutate_adjacent(sequence, n_mutations=1):
    """This was modified by me because it was broken"""
    assert n_mutations <= len(sequence)
    new_sequence = list(sequence)
    nts = ("A", "C", "G", "T")
    ix_to_mutate = np.random.choice(len(new_sequence) - n_mutations + 1)

    for j in range(n_mutations):
        possible_bases = list(nts)
        possible_bases.remove(new_sequence[ix_to_mutate + j])
        new_sequence[int(ix_to_mutate + j)] = \
            np.random.choice(possible_bases, size=1)[0]
    new_sequence = "".join([str(x) for x in new_sequence])

    # Test
    n_diff = 0
    for i in range(len(sequence)):
        if sequence[i] != new_sequence[i]:
            n_diff += 1
    assert n_diff == n_mutations
    return new_sequence


def calc_dtw(scrappie_df_1, scrappie_df_2, warp_width=0.1):
    """Given two lists of currents (the 'current column' of the scrappie df),
     calculate their dtw distance"""
    _, dtw_dist = ucrdtw(scrappie_df_1, scrappie_df_2, warp_width, False)
    dtw_dist = np.float32(dtw_dist)
    return dtw_dist


def calc_all_dtw(seq1, seq2, warp_width=0.1):
    """Given two sequences, calculate DTW distances for all 4 comparisons
    of revcom to normal"""
    rc_seq1 = reverse_complement(seq1)
    rc_seq2 = reverse_complement(seq2)
    all_dists = []
    for fragment_1, fragment_2 in product([seq1, rc_seq1], [seq2, rc_seq2]):
        squig_df1 = SQUIGGLE_DICT[fragment_1]
        squig_df2 = SQUIGGLE_DICT[fragment_2]
        _, dtw_dist = ucrdtw(squig_df1, squig_df2, warp_width, False)
        all_dists.append(dtw_dist)
    dtw_dist = np.array(all_dists, dtype=np.float32)
    return np.min(dtw_dist)

def calc_all_sw(seq1, seq2, cost_fn=None, return_alignment=True):
    """Given two sequences, calculate all 4 SW alignment scores for all
    orientations of the strands"""
    if cost_fn is None:
        cost_fn = {"match": 1, "mismatch": -1, "gap": -4}
    rc_seq1 = reverse_complement(seq1)
    rc_seq2 = reverse_complement(seq2)
    highest_sw = 0
    best_alignment = None
    for fragment_1, fragment_2 in product([seq1, rc_seq1], [seq2, rc_seq2]):
        _, sw, a, _, _ = s_w(fragment_1, fragment_2, cost_fn)
        if sw > highest_sw:
            highest_sw = sw
            best_alignment = a
    if return_alignment:
        return highest_sw, best_alignment
    return highest_sw

def load_squiggles_from_csv(file_path):
    """Get CSV of all kmers and their squiggles and convert it to dict"""
    df = pd.read_csv(file_path, index_col=0, header=None)
    df = df.transpose()
    return df.to_dict('list')


def initialize_distance(sequences):
    """Given sequences and dictionary with all squiggles, create a distance
    matrix that shows DTW distance"""
    n_seqs = len(sequences)
    D = np.zeros((n_seqs, n_seqs))
    for j, seq1 in enumerate(sequences):
        for i, seq2 in enumerate(sequences):
            if i >= j:
                continue
            else:
                dtw_dist = calc_all_dtw(seq1, seq2)
                print(round(j / n_seqs, 2), seq1, seq2, dtw_dist)
                D[i, j] = dtw_dist
                D[j, i] = dtw_dist
    return D


def recalc_distance(D, row_ix, new_seq, seq_list):
    """Update distance matrix d for new squiggle
    """
    D_update = D.copy()
    for j in range(D.shape[0]):
        if j == row_ix:
            continue
        kmer_to_compare = seq_list[j]
        dtw_dist = calc_all_dtw(new_seq, kmer_to_compare)
        D_update[j, row_ix] = dtw_dist
        D_update[row_ix, j] = dtw_dist
    return D_update


def calculate_new_distances(new_seq: str, sequences: list, ignore_ix=None,
                            sw_cost_fn=None):
    """Given a new sequence, calculate smith waterman distances to other sequences,
    but ignores seqs with index in ignore_ix"""
    if sw_cost_fn is None:
        sw_cost_fn = {"match": 1, "mismatch": -1, "gap": -4}
    d_sw = np.zeros(len(sequences))
    for seq_j in range(len(sequences)):
        if ignore_ix is not None and seq_j in ignore_ix:
            continue
        _, sw, _, _, _ = s_w(new_seq, sequences[seq_j], cost_fn=sw_cost_fn)
        d_sw[seq_j] = sw
    return d_sw


def check_improvement(d_old, d_new):
    # d is a single row of D
    identity_ix = np.where(np.logical_and(d_old == 0, d_new == 0))
    d_old = np.delete(d_old, identity_ix)
    d_new = np.delete(d_new, identity_ix)
    min_improved = min(d_old) < min(d_new)
    avg_improved = np.mean(d_old) < np.mean(d_new)
    return min_improved and avg_improved


def make_random_seq(length=40):
    """Generates dna string of nonrepeating bases"""

    def norepeat_DNA():
        seq = ["A", "C", "G", "T"]
        last_bp = np.random.choice(seq)
        yield last_bp
        while True:
            next_bp = np.random.choice(seq)
            if next_bp != last_bp:
                last_bp = next_bp
                yield next_bp

    g = norepeat_DNA()
    seq = ""
    for _ in range(length):
        seq += next(g)
    return seq


def all_sw_comparisons(seq: str , all_seqs: list, seq_i, sw_costs):
    """Perform SW for a sequence against a list of other sequences.
    Do this for all 4 needed orientations: forward to forward,
    forward to reverse complement, reverse complement to forward, 
    and reverse complement to reverse complement.
    
    :param seq: new sequence to compare against all sequences
    :param all_seqs: list of all sequences to compare the seq against
    :param seq_i: index of new sequence
    :param sw_costs: dictionary of smith-waterman alignment costs
    :return: 
    """
    # Reverse complements of sequences
    rc_seq = reverse_complement(seq)
    rc_starting_seqs = [reverse_complement(seq) for seq in all_seqs]

    d_sw1 = calculate_new_distances(seq, all_seqs,
                                    ignore_ix=[seq_i],
                                    sw_cost_fn=sw_costs)
    d_sw2 = calculate_new_distances(seq, rc_starting_seqs,
                                    ignore_ix=[seq_i],
                                    sw_cost_fn=sw_costs)
    d_sw3 = calculate_new_distances(rc_seq, all_seqs,
                                    ignore_ix=[seq_i],
                                    sw_cost_fn=sw_costs)
    d_sw4 = calculate_new_distances(rc_seq, rc_starting_seqs,
                                    ignore_ix=[seq_i],
                                    sw_cost_fn=sw_costs)

    all_distances = np.stack((d_sw1, d_sw2, d_sw3, d_sw4), axis=1)
    max_distances = np.amax(all_distances, axis=1)
    return max_distances


def initialise_kmers(kmer_size, lib_size, match_score, mismatch_score,
                     gap_score, max_sw_sim_init, max_tries=200):
    """Create kmer library where kmers are sufficiently unique

    :param kmer_size: Length of a kmer (we used 8)
    :param lib_size: Total number of different kmers to output
    :param match_score: Smith waterman match score
    :param mismatch_score: Smith waterman mismatch score
    :param gap_score: Smith waterman gap score
    :param max_sw_sim_init: Maximum smith waterman similarity at initialisation
    :param max_tries: Maximum attempts per kmer at generating a valid one
    :return: list of kmers in the library
    """

    sw_dict = {"match": match_score,
               "mismatch": mismatch_score,
               "gap": gap_score}

    # Create initial kmers here
    starting_seqs = [make_random_seq(kmer_size) for _ in range(lib_size)]
    new_starting_seqs = starting_seqs[:]
    for seq_i, seq in enumerate(starting_seqs):
        # Check if it is too similar to an other seq
        too_similar = False
        d_sw = all_sw_comparisons(seq, new_starting_seqs, seq_i, sw_dict)
        if np.max(d_sw) > max_sw_sim_init:
            too_similar = True

        valid, reasons = check_valid_seq(seq)
        print(seq_i, valid, too_similar, reasons)
        tries = 0
        new_seq = seq
        while (not valid or too_similar) and tries < max_tries:
            too_similar = True
            tries += 1
            new_seq = make_random_seq(kmer_size)
            if new_seq in new_starting_seqs or reverse_complement(
                    new_seq) in new_starting_seqs:
                valid = False
                print('Already present')
                continue
            valid, _ = check_valid_seq(new_seq)

            d_sw = all_sw_comparisons(new_seq, new_starting_seqs, seq_i,
                                      sw_dict)

            if np.max(d_sw) <= max_sw_sim_init:
                too_similar = False
        else:
            print(f'Performed {tries} tries')
            new_starting_seqs[seq_i] = new_seq
    starting_seqs = new_starting_seqs[:]
    return starting_seqs, d_sw


def initialise_sw_distances(starting_seqs, match_score,
                            mismatch_score, gap_score, sw_print_cutoff):
    """

    :param starting_seqs:
    :param match_score:
    :param mismatch_score:
    :param gap_score:
    :param sw_print_cutoff:
    :return:
    """
    sw_dict = {"match": match_score,
               "mismatch": mismatch_score,
               "gap": gap_score}
    lib_size = len(starting_seqs)
    D_sw = np.zeros((lib_size, lib_size))
    D_sw_flat = []
    for j in range(lib_size):
        for i in range(lib_size):
            if i >= j:
                continue
            else:
                seq1 = starting_seqs[j]
                seq2 = starting_seqs[i]
                sw, a = calc_all_sw(seq1, seq2, cost_fn=sw_dict,
                                    return_alignment=True)
                D_sw[i, j] = sw
                D_sw[j, i] = sw
                D_sw_flat.append(sw)
                if sw >= sw_print_cutoff:  # Print seqs that are similar to inspect manually, value of 7 seems little to high, too
                    print(sw)
                    print(i, j)
                    print()
                    print(a)
                    print()
    return D_sw, D_sw_flat


def evaluate_candidate_seq(new_seq, current_seqs, seq_i, current_D,
                           max_sw_sim, gap_score, match_score,
                           mismatch_score, outer_10_percent):
    """Given a single new sequence, evaluate if it is an improvement compared
    to current seq

    :param new_seq: candidate new sequence
    :param current_seqs: list of current sequences
    :param seq_i: index of sequence to be replaced
    :param current_D: current DTW distance matrix
    :param max_sw_sim: max allowed smith_waterman alignment
    :param gap_score: what is says on the tin
    :param match_score: see above
    :param mismatch_score: see above
    :param outer_10_percent: if and what outer 10 percent to remove
    :return:
    """
    ignore_ix = [seq_i]
    d_sw = all_sw_comparisons(new_seq, current_seqs, ignore_ix, {
        "match": match_score,
        "mismatch": mismatch_score,
        "gap": gap_score})

    if np.max(d_sw) > max_sw_sim:
        print("Too similar!", np.max(d_sw))
        return None

    valid, reasons = check_valid_seq(new_seq,
                                     remove_if_outer10_percent=outer_10_percent)
    if not valid:
        print('Invalid:', reasons)
        return None
    new_D = recalc_distance(current_D, seq_i, new_seq,
                            current_seqs)
    improved = check_improvement(current_D[seq_i, :],
                                 new_D[seq_i, :])
    if improved:
        return new_D, new_seq, d_sw


HOMOPOLYMERS = ["AAAAA", "TTTTT", "GGGG", "CCCC"]


def main(args):
    if not os.path.exists(args.out_directory):
        os.mkdir(args.out_directory)

    global OUTER10_PERCENT
    if args.remove_outer10_percent:
        # TODO this is a temp hardcoded filepath
        with open('/lustre/BIF/nobackup/noord087/mscthesis/9mer_count_in_16s/outer10kmers.json', 'r') as f:
            f = f.read()
            OUTER10_PERCENT = json.loads(f)

    global SQUIGGLE_DICT
    SQUIGGLE_DICT = load_squiggles_from_csv(args.squiggle_dict)

    # Set parameters
    lib_size = args.library_size
    kmer_size = 9
    # Maximum smith waterman alignment score at kmer initialisation step
    max_sw_sim_init = 7
    # Alignment parameters
    match_score = 1
    mismatch_score = -1
    gap_score = -4
    if not args.kmer_seqs:
        print('Creating starting sequences')
        starting_seqs, d_sw = initialise_kmers(kmer_size, lib_size, match_score,
                                               mismatch_score, gap_score,
                                               max_sw_sim_init, max_tries=200)
        assert np.max(d_sw) < kmer_size, "Starting sequences contain duplicates"
        del d_sw  # Making sure we do not use this variable again
        print('Creating initial DTW distance matrix')
        print(starting_seqs)
    else:
        #  Initial sequences provided, drop first column
        starting_seqs = list(pd.read_csv(args.kmer_seqs, nrows=0))[1:]
        print(f'{len(starting_seqs)} kmers loaded successfully')
    D = initialize_distance(starting_seqs)

    # Save to CSV here
    distance_df = pd.DataFrame(D)
    distance_df.columns = starting_seqs
    distance_df.index = starting_seqs
    distance_df.to_csv(os.path.join(args.out_directory,
                                    f'InitialDistancesOf{kmer_size}mers.csv'))

    print('Calculating all smith waterman distances')
    D_sw, D_sw_flat = initialise_sw_distances(starting_seqs, match_score,
                                              mismatch_score, gap_score,
                                              sw_print_cutoff=7)



    # # Uncomment to save images
    # linkage = hc.linkage(sp.distance.squareform(distance_df), method='average')
    # sns_plot = sns.clustermap(distance_df, row_linkage=linkage,
    #                           col_linkage=linkage)
    # sns_plot.savefig(os.path.join(args.out_directory,
    #                               'Cluster_DTW_distances_of_initial_kmers.png'))
    #
    # linkage = hc.linkage(sp.distance.squareform(D_sw+1, checks=False),
    #                      method='complete')
    # plot = sns.clustermap(D_sw+1, row_linkage=linkage, col_linkage=linkage)
    # plot.savefig(os.path.join(args.out_directory,
    #                           'Cluster_SW_distances_of_initial_kmers.png'))

    # Parameters TODO: make these not be hardcoded, but for now they work
    n_rounds = 5  # make  5 or more?
    rounds_per_mutation_no = int(n_rounds/5)
    mutations_per_round = np.repeat(np.array([5, 4, 3, 2, 1]),
                                    rounds_per_mutation_no)
    max_sw_sim = 7  # Throw away new seq if it scores above this score
    max_tries = 30  # How many attempts to retry improving per sequence

    seqs_after_rounds = [starting_seqs]
    D_after_rounds = [D]
    D_sw_after_rounds = [D_sw]
    triu = np.triu_indices(n=len(starting_seqs), k=1)
    print("Starting summary:\nmin:  %0.4f\nmean: %0.4f"
          % (np.min(D[triu].flatten()), np.mean(D[triu].flatten())))

    # # Get indices only of seqs that are either extremely abundant or extremely
    # # rare
    # indices_outer10 = [i for i, s in enumerate(starting_seqs)
    #                    if s in OUTER10_PERCENT]

    for round_i in range(n_rounds):
        # How many nucleotides to change per mutation
        mutation_no = mutations_per_round[round_i]
        print("============== Round %d ==============" % round_i)
        current_seqs = seqs_after_rounds[round_i][:]
        current_D = D_after_rounds[round_i].copy()
        current_D_sw = D_sw_after_rounds[round_i].copy()

        # # Determine bottom 10 percent
        # duplicate_D = current_D.copy()
        # np.fill_diagonal(duplicate_D, kmer_size)
        # min_dtw_dists = np.amin(duplicate_D, axis=0)
        # # Indices of worst 10 percent
        # worst_10_percent_idx = np.argpartition(min_dtw_dists,
        #                                        int(lib_size * 0.1))[:int(lib_size*0.1)]
        for count, seq_i in enumerate(np.random.choice(range(len(starting_seqs)),
                                      size=len(starting_seqs),
                                      replace=False)):
            print(f'Progress: {((count / len(starting_seqs)) * 100):.2f} percent')
            assert len(current_seqs[seq_i]) == kmer_size

            # # Temporary: only modify seqs that are in outer10 percent
            # if seq_i not in indices_outer10:
            #     continue

            candidates = [mutate_adjacent(current_seqs[seq_i],
                                      n_mutations=mutation_no) for _ in range(max_tries)]

            parallel_args = [(new_seq, current_seqs, seq_i, current_D,
                              max_sw_sim, gap_score, match_score,
                              mismatch_score, args.remove_outer10_percent)
                             for new_seq in candidates]

            with multiprocessing.Pool(max_tries) as p:
                good_candidates = p.starmap(evaluate_candidate_seq, parallel_args)

            if all(i is None for i in good_candidates):
                print('No improvement')
            else:
                for candidate in good_candidates:
                    if candidate:
                        new_D, new_seq, d_sw = candidate
                        print(f"Improved {current_seqs[seq_i]} to {new_seq}")
                        current_seqs[seq_i] = new_seq
                        current_D = new_D.copy()
                        current_D_sw[seq_i, :] = d_sw
                        current_D_sw[:, seq_i] = d_sw
                        break

        seqs_after_rounds.append(current_seqs)
        #     scrappie_after_rounds.append(current_dfs)
        D_after_rounds.append(current_D.copy())
        D_sw_after_rounds.append(current_D_sw.copy())
        print("Improvement summary:\nmin:  %0.4f\nmean: %0.4f" % (
            np.min(current_D[triu].flatten()),
            np.mean(current_D[triu].flatten())))

        df = pd.DataFrame(current_D)
        df.columns = current_seqs
        df.index = current_seqs
        df.to_csv(os.path.join(args.out_directory,
                               f'1500libraryOf{kmer_size}mers_{round_i}'
                               f'_rounds_DTW.csv'))

        df2 = pd.DataFrame(current_D_sw)
        df2.columns = current_seqs
        df2.index = current_seqs
        df2.to_csv(os.path.join(args.out_directory,
                                f'1500libraryOf{kmer_size}mers_{round_i}'
                                f'_rounds_SW.csv'))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Perform evolutionary '
                                                 'algorithm to generate '
                                                 'library of kmers')
    parser.add_argument('--squiggle-dict', required=True,
                        help='Path to CSV containing all simulated squiggles')
    parser.add_argument('--library-size', required=True, type=int,
                        help='Number of different k-mers to generate')
    parser.add_argument('-o', '--out-directory', required=True,
                        help='Output directory')
    parser.add_argument('-s', '--kmer-seqs', required=False,
                        help='Instead of generating random initial kmer sequences, use seqs extracted from distance csv file (output by previous run of this script).'
                             'This argument should be the path to such a csv file.')
    parser.add_argument('--remove-outer10-percent', action='store_true',
                        default=False, help='If this flag is provided, remove '
                                            'kmers that have an abundance in '
                                            'the outer 10%% in 16s rRNA '
                                            'sequences')

    args = parser.parse_args()
    main(args)
